# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_optimization.quantum_annealing.ipynb (unless otherwise specified).

__all__ = ['QuantumAnnealer', 'SUPPORTED_TASKS', 'QA_LOCAL']

# Cell

import networkx as nx
import warnings

# dwave imports
import dwave_networkx as dnx
from dwave.system.composites import EmbeddingComposite

# local imports
from .base import QuantumOptimizer, DEFAULT_LOCAL_SIMULATOR


SUPPORTED_TASKS = {'maximum_clique': dnx.maximum_clique,
                  'minimum_vertex_cover': dnx.min_vertex_cover,
                   'minimum_weighted_vertex_cover' :dnx.min_weighted_vertex_cover,
                   'maximum_independent_set': dnx.maximum_independent_set,
                   'maximum_weighted_independent_set' : dnx.maximum_weighted_independent_set,
                   'maximum_cut' : dnx.maximum_cut,
                   'weighted_maximum_cut' : dnx.weighted_maximum_cut,
                   'traveling_salesperson' : dnx.traveling_salesperson
                  }
QA_LOCAL = 'dwave-neal'

class QuantumAnnealer:

    """quantum-based combinatorial optimization with Amazon Braket


    Usage:

    ```python
    >>> qo = QuantumAnnealer(g,
                            task='maximum_clique',
                            device_name='local',     # uses local simulator solver
                            device_arn='device_arn', # only needed if not local
                            s3_folder='s3_folder')   # only needed if not local
        qo.fit()
        results = qo.results()
    ```

    **Parameters:**

    * **g** : a networkx Graph object
    * **task** : one of {'maximum_clique'}
    * **device_name** : one of {'braket.local.qubit', 'braket.aws.qubit'}
    * **device_arn** : Device ARN. Only required if name != 'local'.
    * **s3_folder** : S3 folder. Only required if name != 'local'.
    """
    def __init__(self, g, task=None,
                 local=True, device_arn=None, s3_folder=None):
        """
        constructor
        """
        # error checks
        if task not in SUPPORTED_TASKS: raise ValueError(f'task {task} is not supported. ' +\
                                                         f'Supported tasks: {list(SUPPORTED_TASKS.keys())}')
        if not local and (device_arn is None or s3_folder is None):
            raise ValueError('device_arn and s3_folder are required if using managed AWS device')
        if local and (device_arn is not None or s3_folder is not None):
            warnings.warn('local=True is being ignored since device_arn and s3_folder exist. ')
        if not isinstance(g, nx.Graph): raise ValueError('g must be instance of networkx.Graph')

        # input vars
        self.g = g
        self.task = task
        self.local = local
        self.device_arn = device_arn
        self.s3_folder = s3_folder


        # computed vars
        self._last_result = None
        self._fit_called = False

    @staticmethod
    def supported_tasks():
        for task in SUPPORTED_TASKS:
            print(task)

    def fit(self, verbose=1):
        """
        Approximate a solution to given task.
        Simulated Annealing is used when `QuantumAnnealer.local=True`.
        Quantum Annealing is used when `QuantumAnnealer.local=False`.
        """

        # setup sampler
        if self.local:
            if self.g.number_of_nodes() < 18:
                from dimod.reference.samplers import ExactSolver
                sampler = ExactSolver()
            else:
                import neal
                sampler = neal.SimulatedAnnealingSampler()
        else:
            bracket_sampler = BraketSampler(self.s3_folder, self.device_arn)
            sampler = EmbeddingComposite(braket_sampler)

        # generate approximation
        kwargs = {}
        if 'weighted' in 'self.task': kwargs['weight'] = 'weight'
        apx_fn = SUPPORTED_TASKS[self.task]
        if self.local:
            result = apx_fn(self.g, sampler, **kwargs)
        else:
            kwargs['resultFormat'] = 'HISTOGRAM'
            #result = apx_fn(self.g, sampler, resultFormat="HISTOGRAM")
            result = apx_fn(self.g, sampler, **kwargs)


        self._last_result = result
        self._fit_called = True
        return self

    def results(self, **kwargs):
        """
        Return approximated solution
        """
        if not self._fit_called: raise Exception('The fit method must be called first.')
        if 'return_probs' in kwargs and kwargs['return_probs']:
            warnings.warn('return_probs not currently supported in QuantumAnnealer, '+\
                          'so returning None for second return value.')
            return self._last_result, None
        else:
            return self._last_result


    def plot_samples(self, probs):
        """
        Plot sample for toy problems for testing purposes.
        """
        raise NotImplemented('QuantumAnnealer does not currently support plot_samples.')



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/98_utils.ipynb (unless otherwise specified).

__all__ = ['add_metrics_to_log', 'log_to_message', 'ProgressBar']

# Cell
#-------------------------------------------------------------------------------
# The ProgressBar, log_to_message, and add_metrics_to_log functions
# support Keras-style training progress in Jupyter notebooks.
# Code Reference:  https://github.com/henryre/pytorch-fitmodule
#-------------------------------------------------------------------------------
import numpy as np
import sys
def add_metrics_to_log(log, metrics, y_true, y_pred, prefix=''):
    for metric in metrics:
        q = metric(y_true, y_pred)
        log[prefix + metric.__name__] = q
    return log


def log_to_message(log, precision=4, omit=[]):
    fmt = "{0}: {1:." + str(precision) + "f}"
    fmt_fn = lambda k,v: "%s:%s" % (k,v) if k == 'iteration' else fmt.format(k,v)
    sep = "  "
    #return sep+sep.join(fmt.format(k, v) for k, v in log.items() if k not in omit)
    return sep+sep.join(fmt_fn(k, v) for k, v in log.items() if k not in omit)



class ProgressBar(object):
    """Progress bar"""

    def __init__(self, n, length=40):
        # Protect against division by zero
        self.n      = max(1, n)
        self.nf     = float(n)
        self.length = length
        # Precalculate the i values that should trigger a write operation
        self.ticks = set([round(i/100.0 * n) for i in range(101)])
        self.ticks.add(n-1)
        self.bar(0)

    def bar(self, i, message=""):
        """Assumes i ranges through [0, n-1]"""
        if i in self.ticks:
            b = int(np.ceil(((i+1) / self.nf) * self.length))
            sys.stdout.write("\r[{0}{1}] {2}%\t{3}".format(
                "="*b, " "*(self.length-b), int(100*((i+1) / self.nf)), message
            ))
            sys.stdout.flush()

    def close(self, message=""):
        # Move the bar to 100% before closing
        self.bar(self.n-1)
        sys.stdout.write("{0}\n\n".format(message))
        sys.stdout.flush()


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"QuantumOptimizer": "01_optimization.base.ipynb",
         "DEFAULT_LOCAL_SIMULATOR": "01_optimization.base.ipynb",
         "QuantumAnnealer": "02_optimization.quantum_annealing.ipynb",
         "SUPPORTED_TASKS": "03_optimization.qaoa.ipynb",
         "QA_LOCAL": "02_optimization.quantum_annealing.ipynb",
         "QAOA": "03_optimization.qaoa.ipynb",
         "QAOA_LOCAL": "03_optimization.qaoa.ipynb",
         "add_metrics_to_log": "98_utils.ipynb",
         "log_to_message": "98_utils.ipynb",
         "ProgressBar": "98_utils.ipynb"}

modules = ["core.py",
           "optimization/base.py",
           "optimization/quantum_annealing.py",
           "optimization/qaoa.py",
           "utils.py"]

doc_url = "https://amaiya.github.io/quixotic/"

git_url = "https://github.com/amaiya/quixotic/tree/main/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_optimization.ipynb (unless otherwise specified).

__all__ = ['QuantumOptimizer', 'SUPPORTED_TASKS', 'DEFAULT_LOCAL']

# Cell
import pennylane as qml
import numpy as np
from matplotlib import pyplot as plt
import networkx as nx
import warnings

SUPPORTED_TASKS = {'maximum_clique': qml.qaoa.max_clique,
                   'minimum_vertex_cover': qml.qaoa.min_vertex_cover}
DEFAULT_LOCAL = 'braket.local.qubit'

class QuantumOptimizer:
    """quantum-based combinatorial optimization with Amazon Braket


    Usage:

    ```python
    >>> qo = QuantumOptimizer(g,
                              task='maximum_clique',
                              device_name='braket.local.qubit', # default braket local simulator
                              device_arn='device_arn',
                              s3_folder='s3_folder')
        qo.fit()
        results = qo.results()
    ```

    **Parameters:**

    * **g** : a networkx Graph object
    * **task** : one of {'maximum_clique'}
    * **device_name** : one of {'braket.local.qubit', 'braket.aws.qubit'}
    * **device_arn** : Device ARN. Only required if name != 'braket.local.simulator'.
    * **s3_folder** : S3 folder. Only required if name != 'braket.local.simulator'.
    * **n_layers**  : number of layers. default:4
    * **constrained**  : default:False
    * **params**  : user-supplied initialization of params.  If None, a default random initialization is used.


    """
    def __init__(self, g, task=None, n_layers=4, constrained=False,
                 device_name='braket.local.qubit', device_arn=None, s3_folder=None,
                 params=None):
        """
        constructor
        """
        # error checks
        if task not in SUPPORTED_TASKS: raise ValueError(f'task {task} is not supported')
        if device_name != DEFAULT_LOCAL and (device_arn is None or s3_folder is None):
            raise ValueError('device_arn and s3_folder are required if using managed AWS device')
        if not isinstance(g, nx.Graph): raise ValueError('g must be instance of networkx.Graph')

        # input vars
        self.g = g
        self.task = task
        self.device_name = device_name
        self.device_arn = device_arn
        self.s3_folder = s3_folder
        self.n_layers = n_layers
        self.constrained=constrained

        # computed vars
        self.wires = self.g.number_of_nodes()
        np.random.seed(1967)
        self.params = self._default_params() if params is None else params
        self.circuit, self.cost_function = self._prepare()
        self._fit_called = False

    def _default_params(self):
        #return np.random.uniform(size=[2, n_layers])
        return np.full((2,self.n_layers), 0.5)


    def _device(self, shots=None):
        kwargs = {}
        if self.device_name != DEFAULT_LOCAL:
            kwargs['device_arn'] = self.device_arn
            kwargs['s3_destination_folder'] = self.s3_folder
        if shots is not None:
            kwargs['shots'] = shots
        dev = qml.device(self.device_name, wires=self.wires, **kwargs)
        return dev

    def _prepare(self):

        # qaoa layer
        cost_h, mixer_h = SUPPORTED_TASKS[self.task](self.g, constrained=self.constrained)
        def qaoa_layer(gamma, alpha):
            qml.qaoa.cost_layer(gamma, cost_h)
            qml.qaoa.mixer_layer(alpha, mixer_h)

        # build circuit
        def circuit(params, **kwargs):
            for i in range(self.wires):  # Prepare an equal superposition over all qubits
                qml.Hadamard(wires=i)
            qml.layer(qaoa_layer, self.n_layers, params[0], params[1])

        # build cost function
        cost_function = qml.ExpvalCost(circuit, cost_h, self._device(), optimize=True)

        return circuit, cost_function


    def fit(self, iterations=10, verbose=1):
        """
        approximate a solution to given task
        """
        # progress bar
        from collections import OrderedDict
        from .utils import add_metrics_to_log, log_to_message, ProgressBar
        if verbose:
            print(f"Optimize for {iterations} iterations...")
            pb = ProgressBar(iterations)
        log = OrderedDict()

        # optimize
        optimizer = qml.GradientDescentOptimizer()
        for i in range(iterations):
            self.params = optimizer.step(self.cost_function, self.params)
            cost_eval = self.cost_function(self.params)
            #if verbose:
            #    print(f"Completed iteration {i + 1}, cost function:", cost_eval)
            log['cost'] = cost_eval
            log['iteration'] = i+1
            if verbose:
                pb.bar(i, log_to_message(log))
        self._fit_called = True
        return

    def results(self, shots = 100000):
        """
        Sample and return results.
        Returns a tuple of the form: `(nodes, probs)`, where `nodes` are the list
        of vertices comprising the most likely solution and `probs` are the raw solution probabilities.
        """

        if not self._fit_called: raise Exception('The fit method must be called first.')

        dev = self._device(shots=shots)

        @qml.qnode(dev)
        def samples(params):
            self.circuit(params)
            return [qml.sample(qml.PauliZ(i)) for i in range(self.wires)]

        # compute probabilities
        from collections import Counter
        s = samples(self.params).T
        s = (1 - s.numpy()) / 2
        s = map(tuple, s)
        counts = Counter(s)
        indx = np.ndindex(*[2] * self.wires)
        probs = {p: counts.get(p, 0) / shots for p in indx}

        # extract nodes
        import operator
        result_bool = max(probs.items(), key=operator.itemgetter(1))[0]
        nodes = [i for i, b in enumerate(result_bool) if b]

        return nodes, probs


    def plot_samples(self, probs):
        """
        Plot sample for toy problems for testing purposes.
        """

        import matplotlib.pyplot as plt

        plt.style.use("seaborn")
        labels = ["{0:{fill}8b}".format(i, fill='0') for i in range(len(probs))]

        plt.bar(range(2 ** self.wires), probs.values())
        plt.xticks([i for i in range(len(probs))], labels, rotation='vertical', size=12)
        plt.yticks(size=12)

        plt.xlabel("Sample", size=20)
        plt.ylabel("Probability", size=20)

        fig = plt.gcf()
        fig.set_size_inches(16, 8)
        plt.show()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_optimization.base.ipynb (unless otherwise specified).

__all__ = ['QuantumOptimizer', 'DEFAULT_LOCAL_SIMULATOR']

# Cell

DEFAULT_LOCAL_SIMULATOR = 'local'

from abc import ABC, abstractmethod
class QuantumOptimizer(ABC):
    """
    ```
    Abstract class to preprocess data
    ```
    """
    @abstractmethod
    def fit(self):
        pass

    @abstractmethod
    def results(self, shots=512):
        pass



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_core.ipynb (unless otherwise specified).

__all__ = []

# Cell
from .optimization.quantum_annealing import QuantumAnnealer
from .optimization.qaoa import QAOA